#!/bin/bash

#usage, if exit with error then use "usage 1", or "usage 0"
#filename: ${type}-${device}-${mode}-${version}-${tester}-${timestamp}
usage()
{
    echo "Usage: $(basename $0) --log-dir <log dir> --output-dir <out-dir> --type [end|reboot] --dev <dev> --mode [test tag] --ver <ver> --tester <tester> --date-from <yymmddHHMM> --date-to <yymmddHHMM> [--help|-h]"
    echo "  --type [end|reboot], filter log by type, the type was manual defined"
    echo "  --dev <dev>, the string get from device by 'getprop ro.product.device'"
    echo "  --mode [release|custom*], the config mode, such as release, daily, local and so on"
    echo "  --ver <ver>, the string get from device by 'getprop ro.build.version.incremental'"
    echo "  --tester <tester>, the string get from tester machine by 'cat /etc/hostname'"
    echo "  --date-from <yymmddHHMM>"
    echo "  --date-to <yymmddHHMM>, filter the logs by date, the format must be full of yymmddHHMM"
    echo ""
    echo "  <dev>, <mode>, <tester>, can use wildcard as shell, for example: ? and *"
    exit $1
}

#options
log_dir=
report_dir=
log_type=
dev=
ver=
mode=
tester=
date_from=
date_to=

#get options
while [ $# -gt 0 ]
do
    case $1 in
    --log-dir)
        shift
        [ -d "$1" ] || usage 1
        log_dir=$1
        ;;
    --output-dir)
        shift
        report_dir=$1
        ;;
    --type)
        shift
        log_type=$1
        [ "$log_type" == "end" ] || [ "$log_type" == "reboot" ] || usage 1
        ;;
    --dev)
        shift
        dev=$1
        ;;
    --ver)
        shift
        ver=$1
        ;;
    --mode)
        shift
        mode=$1
        ;;
    --tester)
        shift
        tester=$1
        ;;
    --date-from)
        shift
        [[ "$1" == *[!0-9]* ]] && usage 1
        date_from=$1
        ;;
    --date-to)
        shift
        [[ "$1" == *[!0-9]* ]] && usage 1
        date_to=$1
        ;;
    --help | -h)
        usage 0
        ;;
    -*)
        echo "Unrecognized option $1"
        usage 1
        ;;
    *)
        break
        ;;
    esac

    shift
done

[ -n "$log_dir" ] || usage 1 
log_dir=$(cd $log_dir;pwd)

[ -n "$report_dir" ] || usage 1
mkdir -p $report_dir
report_dir=$(cd $report_dir;pwd)

[ -n "$log_type" ] || log_type="*"

if [ -n "$dev" ]
then
    dev=$(echo $dev | sed 's/-/_/g')
else
    dev="*"
fi

if [ -n "$ver" ]
then
    ver=$(echo $ver | sed 's/-/_/g')
else
    ver="*"
fi

if [ -n "$mode" ]
then
    mode=$(echo $mode | sed 's/-/_/g')
else
    mode="*"
fi

if [ -n "$tester" ]
then
    tester=$(echo $tester | sed 's/-/_/g')
else
    tester="*"
fi

cat > ${report_dir}/summary << INFO
[log report informations]
log folder    : $log_dir
report folder : $report_dir
log filter    : $log_type-$dev-$mode-$ver-$tester(type-dev-mode-ver-tester)
timestamps    : from ($date_from) to ($date_to)
--------------------
INFO

end_log_file=$(tempfile)
reboot_log_file=$(tempfile)
other_log_file=$(tempfile)

files=mtlog-$log_type-$dev-$mode-$ver-$tester-*.tar.bz2

index_file=$(tempfile)
content_file=$(tempfile)
file_index=0

printf "Index\tLast\tTombs\tMinidmp\tOOM-b2g\n" >> $content_file

cd $log_dir && ls $files | while read file
do
    file_name=${file##*/}
    file_raw=${file_name%%.tar.bz2}
    cur_date=${file_raw##*-}

    if [ -n "$date_from" ]
    then
        [ "$date_from" -lt "$cur_date" ] || continue
    fi

    if [ -n "$date_to" ]
    then
        [ "$cur_date" -le "$date_to" ] || continue
    fi

    cur_type=$(echo $file_raw | cut -d- -f2)

    case "$cur_type" in
    "end")
        echo $cur_date $file_raw >> $end_log_file
        ;;
    "reboot")
        echo $cur_date $file_raw >> $reboot_log_file
        ;;
    *)
        echo $cur_date $file_raw >> $other_log_file
        ;;
    esac

    file_index=$((file_index+1))

    tar -xavf $file -C ${report_dir} ${file_raw}/minidump_summary ${file_raw}/slog_report ${file_raw}/oom-kill-details > /dev/null 2>&1
    echo $file_index $file_raw >> $index_file
    lasttime=$(grep -P "^MonkeyLastTime" ${report_dir}/${file_raw}/slog_report 2>/dev/null | awk -F':' '{print $2}')
    [ -n "$lasttime" ] || lasttime=0
    tombstone_cnt=$(grep -P "^Tombstones Count" ${report_dir}/${file_raw}/slog_report 2>/dev/null | awk -F':' '{print $2}')
    [ -n "$tombstone_cnt" ] || tombstone_cnt=0
    minidump_cnt=$(grep -P "^Count:" ${report_dir}/${file_raw}/minidump_summary 2>/dev/null | awk -F':' '{print $2}')
    [ -n "$minidump_cnt" ] || minidump_cnt=0
    oom_b2g_cnt=$(grep -P "^OOM-Killer Kill B2G Count" ${report_dir}/${file_raw}/slog_report 2>/dev/null | awk -F':' '{print $2}')
    [ -n "$oom_b2g_cnt" ] || oom_b2g_cnt=0
    printf "%d\t%d\t%d\t%d\t%d\n" $file_index $lasttime $tombstone_cnt $minidump_cnt $oom_b2g_cnt >> $content_file
done

[ $(stat $end_log_file -c %s) -gt 0 ] && cat $end_log_file | sort > ${report_dir}/log-end
[ $(stat $reboot_log_file -c %s) -gt 0 ] && cat $reboot_log_file | sort > ${report_dir}/log-reboot
[ $(stat $other_log_file -c %s) -gt 0 ] && cat $other_log_file | sort > ${report_dir}/log-other

cnt_end=$(cat $end_log_file | wc -l)
cnt_reboot=$(cat $reboot_log_file | wc -l)
cnt_other=$(cat $other_log_file | wc -l)
total=$((cnt_end+cnt_reboot+cnt_other))
# data report, give a simple summary of all cr
cat >> ${report_dir}/summary << SUMMARY
$cnt_end(end) + $cnt_reboot(reboot) + $cnt_other(other) = $total(total)
SUMMARY

cat $index_file >> ${report_dir}/summary
echo "--------------------" >> ${report_dir}/summary
cat $content_file >> ${report_dir}/summary
echo "--------------------" >> ${report_dir}/summary
cnt_tombs=$(cat $content_file | grep -P "^[\d\s]+$" | awk '{sum+=$3} END{print sum}')
cnt_minidump=$(cat $content_file | grep -P "^[\d\s]+$" | awk '{sum+=$4} END{print sum}')
cnt_oom_b2g=$(cat $content_file | grep -P "^[\d\s]+$" | awk '{sum+=$5} END{print sum}')
printf "%d\t\t%d\t%d\t(Total)\n" $total $cnt_tombs $cnt_minidump >> ${report_dir}/summary

#end-reboot-other-total-tombstones-minidump
printf "%d\t%d\t%d\t%d\t%d\t%d\t%d\n" $cnt_end $cnt_reboot $cnt_other $total $cnt_tombs $cnt_minidump $cnt_oom_b2g

exit 0
